// Copyright (C) 2000, Luca Padovani <luca.padovani@cs.unibo.it>.
// 
// This file is part of GtkMathView, a Gtk widget for MathML.
// 
// GtkMathView is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// GtkMathView is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with GtkMathView; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// 
// For details, see the GtkMathView World-Wide-Web page,
// http://cs.unibo.it/~lpadovan/mml-widget, or send a mail to
// <luca.padovani@cs.unibo.it>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

// please note that the header file for this module has extension .h,
// since it can be included even by C modules.

#include <assert.h>

#include "Clock.hh"
#include "Parser.hh"
#include "globals.hh"
#include "CCInterface.h"
#include "CharMapper.hh"
#include "DrawingArea.hh"
#include "EntitiesTable.hh"
#include "MathMLDocument.hh"
#include "OperatorDictionary.hh"
#include "RenderingEnvironment.hh"

static MathMLElement*
#ifdef HELM_PROJECT
select_minimum_tree(SelectMode, MathMLElement*, MathMLElement*);
#else
select_minimum_tree(MathMLElement*, MathMLElement*);
#endif

// the real implementation of CCInterface follows. It is hidden
// from outside.
struct _CCInterface {
  GtkMathView*    math_view;
  MathMLDocument* document;
  MathMLElement*  root;

  float           font_size;

  MathMLElement*  select_first;
  MathMLElement*  select_last;
  MathMLElement*  select_root;

  MathMLElement*  clicked;
};

extern "C" void
init_global_data()
{
  //entitiesTable.Load("entities/entities.xml");
  entitiesTable.LoadInternalTable();
  dictionary.Load("/usr/local/share/gtkmathview/dictionary.xml");
  charMapper.Load("/usr/local/share/gtkmathview/fonts.xml");
  dictionary.Load("config/dictionary.xml");
  charMapper.Load("config/fonts.xml");
}

extern "C" CCInterface*
create_interface(GtkMathView* math_view)
{
  assert(math_view != NULL);

  CCInterface* interface = new CCInterface;

  interface->math_view = math_view;
  interface->document = NULL;
  interface->root = NULL;

  interface->font_size = 14.0;

  interface->select_first = interface->select_last = NULL;
  interface->select_root = NULL;
  interface->clicked = NULL;

  return interface;
}

extern "C" void
destroy_interface(CCInterface* interface)
{
  if (interface == NULL) return;
  unload_document(interface);
}

extern "C" gboolean
load_document(CCInterface* interface, const gchar* name)
{
  assert(interface != NULL);
  assert(name != NULL);

  delete interface->document;
  interface->document = NULL;
  interface->root = NULL;

  interface->select_first =
    interface->select_last =
    interface->select_root = NULL;
  interface->clicked = NULL;

#ifdef DEBUG
  Clock perf;
  perf.Start();
#endif
  MathMLParser *parser = new MathMLParser(name);
  interface->document = parser->Parse();
#ifdef DEBUG
  perf.Stop();
  perf.Dump("%%% parsing: ");
#endif
  delete parser;

  if (interface->document != NULL) {
#ifdef DEBUG
    perf.Start();
#endif
    interface->document->Normalize();
#ifdef DEBUG
    perf.Stop();
    perf.Dump("%%% normaliz: ");
#endif

    MathMLElement* root = interface->document->GetRoot();
    assert(root != NULL);

    interface->root = root;

    setup_document(interface);

    return TRUE;
  }

  return FALSE;
}

extern "C" void
setup_document(CCInterface* interface)
{
  assert(interface != NULL);
  assert(interface->root != NULL);

  Clock perf;

  UnitValue size;
  size.Set(interface->font_size, UNIT_PT);

  RenderingEnvironment* env = new RenderingEnvironment();
  env->SetFontSize(size);
#ifdef DEBUG
  perf.Start();
#endif
  interface->root->Setup(env);
#ifdef DEBUG
  perf.Stop();
  perf.Dump("%%% setup: ");
#endif
  delete env;

#ifdef DEBUG
  perf.Start();
#endif
  interface->root->DoBoxedLayout(LAYOUT_MIN);
#ifdef DEBUG
  perf.Stop();
  perf.Dump("%%% m layout:");
#endif

#ifdef DEBUG
  perf.Start();
#endif
  interface->root->DoBoxedLayout(LAYOUT_MAX);
#ifdef DEBUG
  perf.Stop();
  perf.Dump("%%% M layout:");
#endif
}

extern "C" void
unload_document(CCInterface* interface)
{
  assert(interface != NULL);

  delete interface->document;

  interface->document = NULL;
  interface->root = NULL;
  interface->math_view = NULL;
  interface->clicked = NULL;
}

extern "C" void
dump_document(CCInterface* interface)
{
  assert(interface != NULL);

  if (interface->root == NULL) printf("<null document/>\n");
  else interface->root->Dump();
}

extern "C" void
layout_document(CCInterface* interface, gint size)
{
  assert(interface != NULL);

  if (interface->root == NULL) return;
  MathMLElement* root = interface->root;

#ifdef DEBUG
  Clock perf;
  perf.Start();
#endif
  root->DoBoxedLayout(LAYOUT_AUTO, px2sp(size - 10));
  root->SetPosition(px2sp(5), px2sp(5) + root->GetBoundingBox().ascent);
  root->Freeze();
#ifdef DEBUG
  perf.Stop();
  perf.Dump("%%% layout: ");
#endif
}

static void
set_document_dirty_elem(CCInterface* interface, MathMLElement* elem)
{
  assert(interface != NULL);

  if (interface->root == NULL) return;

  BoundingBox box = elem->GetBoundingBox();
  
  // THIS IS ABSOLUTELY WRONG!
  assert(0);

  Rectangle rect;
  rect.x = elem->GetX();
  rect.y = elem->GetY();
  rect.width = box.width;
  rect.height = box.GetHeight();

  interface->root->SetDirty();
  assert(elem->IsDirty());
}

extern "C" void
render_document(CCInterface* interface, gint x, gint y, gint width, gint height)
{
  assert(interface != NULL);
  assert(interface->math_view != NULL);

  if (interface->root == NULL) return;
  MathMLElement* root = interface->root;

  Rectangle rect;
  rect.x      = px2sp(x);
  rect.y      = px2sp(y);
  rect.width  = px2sp(width);
  rect.height = px2sp(height);

  root->SetDirty(&rect);

  GdkPixmap* pixmap = gtk_math_view_get_buffer(interface->math_view);
#ifdef DEBUG
  Clock clock;
  clock.Start();
#endif
  DrawingArea* area = new DrawingArea(GTK_WIDGET(interface->math_view->area), pixmap);

  area->SetTop(float2sp(interface->math_view->top_x),
	       float2sp(interface->math_view->top_y));

  root->Render(*area);
  area->Update(rect);
#ifdef DEBUG
  clock.Stop();
  //  clock.Dump("%%% render: ");
#endif

  delete area;
}

extern "C" void
update_document(CCInterface* interface)
{
  assert(interface != NULL);
  assert(interface->math_view != NULL);

  if (interface->root == NULL) return;
  MathMLElement* root = interface->root;

  GdkPixmap* pixmap = gtk_math_view_get_buffer(interface->math_view);
#ifdef DEBUG
  Clock clock;
  clock.Start();
#endif
  DrawingArea* area = new DrawingArea(GTK_WIDGET(interface->math_view), pixmap);

  area->SetTop(float2sp(interface->math_view->top_x),
	       float2sp(interface->math_view->top_y));

  root->Render(*area);
  area->Update();
#ifdef DEBUG
  clock.Stop();
  clock.Dump("%%% render: ");
#endif

  delete area;
}

extern "C" gint
get_document_width(CCInterface* interface)
{
  assert(interface != NULL);

  if (interface->root == NULL) return 0;

  BoundingBox box = interface->root->GetBoundingBox();

  return sp2ipx(box.width);
}

extern "C" gint
get_document_height(CCInterface* interface)
{
  assert(interface != NULL);

  if (interface->root == NULL) return 0;

  BoundingBox box = interface->root->GetBoundingBox();

  return sp2ipx(box.GetHeight());
}

extern "C" void
set_selection_first(CCInterface* interface, gint x, gint y)
{
  assert(interface != NULL);
  assert(interface->math_view != NULL);

  if (interface->root == NULL) return;

  MathMLElement* selected = interface->root->Inside(px2sp(x) + float2sp(interface->math_view->top_x),
						    px2sp(y) + float2sp(interface->math_view->top_y));

  if (interface->select_root != NULL) {
    interface->select_root->ResetSelected();
    //set_document_dirty_elem(interface, interface->select_root);
  }

  interface->select_first = selected;
  interface->select_last = interface->select_root = NULL;

  update_document(interface);
}

extern "C" void
set_selection_last(CCInterface* interface, gint x, gint y)
{
  assert(interface != NULL);

  if (interface->root == NULL) return;
  if (interface->select_first == NULL) return;

  interface->select_last = interface->root->Inside(px2sp(x) + float2sp(interface->math_view->top_x),
						   px2sp(y) + float2sp(interface->math_view->top_y));

  if (interface->select_last != NULL) {
#ifdef HELM_PROJECT
    MathMLElement* selected = select_minimum_tree(gtk_math_view_get_select_mode(interface->math_view),
						  interface->select_first,
						  interface->select_last);      
#else
    MathMLElement* selected = select_minimum_tree(interface->select_first,
						  interface->select_last);      
#endif

    if (selected != interface->select_root) {
      if (interface->select_root != NULL) {
	interface->select_root->ResetSelected();
	//set_document_dirty_elem(interface, interface->select_root);
      }

      interface->select_root = selected;

      if (interface->select_root != NULL) {
	interface->select_root->SetSelected();
	//set_document_dirty_elem(interface, interface->select_root);
      }

      update_document(interface);
    }
  }
}

extern "C" const gchar*
get_hyperlink(CCInterface* interface, gint x, gint y)
{
  assert(interface != NULL);

  if (interface->root == NULL) return NULL;

  interface->clicked = interface->root->Inside(px2sp(x) + float2sp(interface->math_view->top_x),
					       px2sp(y) + float2sp(interface->math_view->top_y));

  if (interface->clicked == NULL) return NULL;

  const String* link = interface->clicked->GetAttribute(ATTR_HELM_HREF, NULL, false);
  if (link == NULL) return NULL;

  return link->ToStaticC();
}
  
gboolean
is_on_link(CCInterface* interface, gint x, gint y)
{
  return get_hyperlink(interface, x, y) != NULL;
}

static MathMLElement*
#ifdef HELM_PROJECT
select_minimum_tree(SelectMode mode, MathMLElement* first, MathMLElement* last)
#else
select_minimum_tree(MathMLElement* first, MathMLElement* last)
#endif
{
  assert(first != NULL);
  assert(last != NULL);

  if (first != last) {
    unsigned first_depth = first->GetDepth();
    unsigned last_depth  = last->GetDepth();

    while (first_depth > last_depth) {
      first = first->GetParent();
      first_depth--;
    }

    while (last_depth > first_depth) {
      last = last->GetParent();
      last_depth--;
    }

    assert(first_depth == last_depth);

    while (first != NULL && last != NULL && first != last) {
      first = first->GetParent();
      last = last->GetParent();
    }
  }

  assert(first == last);

#ifdef HELM_PROJECT
  if (mode == MATH_VIEW_SELECT_CONTENT) {
    while (first != NULL && !first->IsSet(ATTR_XLINK_XREF)) {
      first = first->GetParent();
    }
  }
#endif

  return first;
}

extern "C" void
set_font_size(CCInterface* interface, guint size)
{
  assert(interface != NULL);
  assert(size > 0);

  interface->font_size = size;
}

extern "C" const gchar*
get_selection(CCInterface* interface)
{
  assert(interface != NULL);
  if (interface->select_root == NULL) return NULL;

  const String* aValue = interface->select_root->GetAttribute(ATTR_XLINK_XREF, NULL, false);
  assert(aValue != NULL);

  return aValue->ToStaticC();
}

extern "C" void
reset_selection(CCInterface* interface)
{
  assert(interface != NULL);

  if (interface->select_root != NULL) {
    interface->select_root->ResetSelected();
    //set_document_dirty_elem(interface, interface->select_root);
  }

  interface->select_first = interface->select_last = NULL;
  interface->select_root = NULL;

  update_document(interface);
}
