// Copyright (C) 2000, Luca Padovani <luca.padovani@cs.unibo.it>.
// 
// This file is part of GtkMathView, a Gtk widget for MathML.
// 
// GtkMathView is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// GtkMathView is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with GtkMathView; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// 
// For details, see the GtkMathView World-Wide-Web page,
// http://cs.unibo.it/~lpadovan/mml-widget, or send a mail to
// <luca.padovani@cs.unibo.it>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <assert.h>

#include "Layout.hh"
#include "Iterator.hh"
#include "frameAux.hh"
#include "operatorAux.hh"
#include "MathMLElement.hh"
#include "MathMLTextNode.hh"

Layout::Layout(scaled width, BreakId id)
{
  totalWidth = width;
  breakability = id;
  bestPenalty = penalty = 0;
  best = NULL;
  SetLineSpacing(0);
}

Layout::~Layout()
{
}

void
Layout::In()
{
  penalty++;
}

void
Layout::Out()
{
  assert(penalty > 0);
  penalty--;
}

int
Layout::GetPenalty(BreakId id) const
{
  switch (id) {
  case BREAK_AUTO:
  case BREAK_NO:
    return MAX_PENALTY;
  case BREAK_BAD:
    return penalty + 1;
  case BREAK_GOOD:
    return penalty;
  default:
    return 0;
  }
}

void
Layout::Append(MathMLFrame* frame, BreakId id)
{
  assert(frame != NULL);
  frame->ResetLast();
  AppendAtom(frame, getFrameBoundingBox(frame, LAYOUT_MAX).width, id);
}

void
Layout::Append(scaled width, BreakId id)
{
  assert(id != BREAK_YES && id != BREAK_INDENT);
  AppendAtom(NULL, width, id);
}

void
Layout::AppendAtom(MathMLFrame* frame, scaled width, BreakId id)
{
  if (id != BREAK_YES && id != BREAK_INDENT && id > breakability) id = breakability;

  Atom* atom = new Atom;
  atom->frame       = frame;
  atom->width       = width;
  atom->penalty     = GetPenalty(id);
  atom->lineSpacing = lineSpacing;

  if (content.GetSize() == 0) NewRow();
  assert(content.GetLast() != NULL);

  content.GetLast()->content.Append(atom);

  availableWidth -= width;

  if (availableWidth < 0 && best != NULL) {
    Row *row = content.GetLast();
    Row *newRow = new Row;

    availableWidth = totalWidth - content.GetLast()->BreakUpTo(best, newRow);

    row->RemoveDiscardableSpacesRight();
    availableWidth += newRow->RemoveDiscardableSpacesLeft();

    FindBreakPoint(newRow);

    content.Append(newRow);
  }

  UpdateBreakPoint(atom);

  if (id == BREAK_YES || id == BREAK_INDENT) {
    NewRow();
    if (id == BREAK_INDENT) Append(INDENT_SPACING, BREAK_NO);
  }
}

void
Layout::UpdateBreakPoint(Atom* atom)
{
  assert(atom != NULL);

  if (atom->penalty < MAX_PENALTY && atom->penalty <= bestPenalty) {
    bestPenalty = atom->penalty;
    best = atom;
  }
}

void
Layout::FindBreakPoint(Row* row)
{
  assert(row != NULL);

  bestPenalty = MAX_PENALTY;
  best = NULL;

  for (Iterator<Atom*> atom(row->content); atom.More(); atom.Next())
    UpdateBreakPoint(atom());
}

void
Layout::NewRow()
{
  if (content.GetSize() > 0) {
    Row* lastRow = content.GetLast();
    assert(lastRow != NULL);
    if (lastRow->content.GetSize() == 0) return;
  }

  Row* row = new Row;
  content.Append(row);
  availableWidth = totalWidth;
  FindBreakPoint(row);
}

void
Layout::SetMinimumLineSpacing(scaled ls)
{
  lineSpacing = scaledMax(lineSpacing, ls);
}

void
Layout::SetLineSpacing(scaled ls)
{
  lineSpacing = scaledMax(0, ls);
}

void
Layout::DoLayout(LayoutId id) const
{
#if 0
  if (content.GetSize() > 0) {
    Row* row = content.GetLast();
    assert(row != NULL);
    row->RemoveDiscardableSpacesRight();
  }
#endif

  for (Iterator<Row*> row(content); row.More(); row.Next()) row()->DoLayout(id, totalWidth);
}

void
Layout::GetBoundingBox(BoundingBox& box, LayoutId id) const
{
  box.Zero();

  bool first = true;
  scaled lastDescent = 0;
  for (Iterator<Row*> i(content); i.More(); i.Next()) {
    assert(i() != NULL);
    BoundingBox rowBox;

    i()->GetBoundingBox(rowBox, id);
    if (first) {
      box = rowBox;
      lastDescent = rowBox.descent;
      first = false;
    } else {
      scaled extraSkip = scaledMax(0, i()->GetMinimumSpacing() - lastDescent - rowBox.ascent);
      box.descent += rowBox.GetHeight() + extraSkip;
      box.width = scaledMax(box.width, rowBox.width);
      lastDescent = rowBox.descent;
    }
  }
}

void
Layout::SetLastBreakability(BreakId id)
{
  if (content.GetSize() == 0) return;

  if (id != BREAK_YES && id != BREAK_INDENT && id > breakability) id = breakability;

  Row* lastRow = content.GetLast();
  assert(lastRow != NULL);
  if (lastRow->content.GetSize() == 0) return;

  Atom* last = lastRow->content.GetLast();
  assert(last != NULL);

  if (id == BREAK_YES || id == BREAK_INDENT) {
    NewRow();
    if (id == BREAK_INDENT) Append(INDENT_SPACING, BREAK_NO);
  } else {
    unsigned penalty = GetPenalty(id);
    last->penalty = penalty;
    UpdateBreakPoint(last);
  }
}

void
Layout::SetPosition(scaled x, scaled y, ColumnAlignId align)
{
  bool first = true;
  scaled lastDescent = 0;
  for (Iterator<Row*> i(content); i.More(); i.Next()) {
    Row* row = i();
    assert(row != NULL);

    BoundingBox rowBox;
    row->GetBoundingBox(rowBox, LAYOUT_AUTO);
    scaled offset = 0;

    switch (align) {
    case COLUMN_ALIGN_LEFT:
      offset = 0;
      break;
    case COLUMN_ALIGN_RIGHT:
      offset = scaledMax(0, totalWidth - rowBox.width);
      break;
    case COLUMN_ALIGN_CENTER:
      offset = scaledMax(0, (totalWidth - rowBox.width) / 2);
      break;
    default:
      assert(IMPOSSIBLE);
      break;
    }

    if (first) {
      row->SetPosition(x + offset, y);
      y += rowBox.descent;
      lastDescent = rowBox.descent;
      first = false;
    } else {
      scaled extraSkip = scaledMax(0, i()->GetMinimumSpacing() - lastDescent - rowBox.ascent);
      row->SetPosition(x + offset, y + rowBox.ascent + extraSkip);
      y += rowBox.GetHeight() + extraSkip;
      lastDescent = rowBox.descent;
    }
  }
}

scaled
Layout::Row::GetWidth(LayoutId id) const
{
  BoundingBox box;
  GetBoundingBox(box, id);
  return box.width;
}

scaled
Layout::Row::GetMinimumSpacing() const
{
  scaled spacing = 0;
  for (Iterator<Atom*> atom(content); atom.More(); atom.Next())
    spacing = scaledMax(spacing, atom()->lineSpacing);

  return spacing;
}

void
Layout::Row::GetBoundingBox(BoundingBox& box, LayoutId id) const
{
  box.Zero();
  for (Iterator<Atom*> atom(content); atom.More(); atom.Next()) {
    assert(atom() != NULL);
    BoundingBox atomBox;
    atom()->GetBoundingBox(atomBox, id);
    box.Append(atomBox);
  }
}

scaled
Layout::Row::RemoveDiscardableSpacesRight()
{
  scaled width = 0;

  while (content.GetSize() > 0 &&
	 content.GetLast() != NULL &&
	 content.GetLast()->IsDiscardable()) {
    Atom* atom = content.RemoveLast();
    assert(atom != NULL);
    width += atom->width;
    delete atom;
  }

  return width;
}

scaled
Layout::Row::RemoveDiscardableSpacesLeft()
{
  scaled width = 0;

  while (content.GetSize() > 0 &&
	 content.GetFirst() != NULL &&
	 content.GetFirst()->IsDiscardable()) {
    Atom* atom = content.RemoveFirst();
    assert(atom != NULL);
    width += atom->width;
    delete atom;
  }

  return width;
}

void
Layout::Row::DoLayout(LayoutId id, scaled totalWidth) const
{
  unsigned nElements = 0;
  scaled fixedWidth = 0;
  scaled elemWidth = 0;
  for (Iterator<Atom*> atom(content); atom.More(); atom.Next()) {
    assert(atom() != NULL);
    if (atom()->IsSpace()) fixedWidth += atom()->width;
    else {
      assert(atom()->frame != NULL);
      if (atom()->frame->IsText()) {
	fixedWidth += getFrameBoundingBox(atom()->frame).width;
      } else {
	assert(atom()->frame->IsElement());
	elemWidth += getFrameBoundingBox(atom()->frame, LAYOUT_MAX).width;
	nElements++;
      }
    }
  }

  bool moreSpace = fixedWidth + elemWidth < totalWidth;
  scaled availableWidth = scaledMax(0, totalWidth - fixedWidth);

  for (Iterator<Atom*> atom(content); atom.More(); atom.Next()) {
    if (atom()->IsFrame() && atom()->frame->IsElement()) {
      MathMLElement* elem = TO_ELEMENT(atom()->frame);
      assert(elem != NULL);

      if (id == LAYOUT_AUTO) {
	if (!isStretchyOperator(elem, STRETCH_VERTICAL)) {
	  scaled availablePerElem = moreSpace ?
	    scaledMax(availableWidth / nElements, elem->GetMaxBoundingBox().width) :
	    availableWidth / nElements;
	  elem->DoBoxedLayout(id, BREAK_GOOD, availablePerElem);
	  availableWidth -= elem->GetBoundingBox().width;
	} else
	  availableWidth -= elem->GetMaxBoundingBox().width;
      } else {
	elem->DoBoxedLayout(id, (id == LAYOUT_MAX) ? BREAK_NO : BREAK_GOOD);
	availableWidth -= elem->GetBoundingBox().width;
      }
      availableWidth = scaledMax(0, availableWidth);

      nElements--;
    }
  }

  if (content.GetSize() > 0) {
    Iterator<Atom*> p(content);
    p.ResetLast();
    while (p.More()) {
      assert(p() != NULL);
      if (!p()->IsSpace()) break;
      p.Prev();
    }

    if (p.More()) {
      assert(p() != NULL);
      assert(p()->frame != NULL);
      p()->frame->SetLast();
    }
  }
}

void
Layout::Row::SetPosition(scaled x, scaled y)
{
  for (Iterator<Atom*> atom(content); atom.More(); atom.Next()) {
    assert(atom() != NULL);
    atom()->SetPosition(x, y);
    x += atom()->GetWidth(LAYOUT_AUTO);
  }
}

scaled
Layout::Row::BreakUpTo(Atom* atom, Row* newRow)
{
  assert(atom != NULL);
  assert(newRow != NULL);

  scaled width = 0;

  while (content.GetSize() > 0 && content.GetLast() != atom) {
    Atom* atom = content.RemoveLast();
    assert(atom != NULL);
    newRow->content.AddFirst(atom);
    width += atom->width;
  }

  return width;
}

scaled
Layout::Atom::GetWidth(LayoutId id) const
{
  BoundingBox box;
  GetBoundingBox(box, id);
  return box.width;
}

void
Layout::Atom::GetBoundingBox(BoundingBox& box, LayoutId id) const
{
  if (frame != NULL) box = getFrameBoundingBox(frame, id);
  else {
    box.Zero();
    box.width = width;
  }
}

void
Layout::Atom::SetPosition(scaled x, scaled y)
{
  if (frame == NULL) return;
  frame->SetPosition(x, y);
}


