// Copyright (C) 2000, Luca Padovani <luca.padovani@cs.unibo.it>.
// 
// This file is part of GtkMathView, a Gtk widget for MathML.
// 
// GtkMathView is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// GtkMathView is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with GtkMathView; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// 
// For details, see the GtkMathView World-Wide-Web page,
// http://cs.unibo.it/~lpadovan/mml-widget, or send a mail to
// <luca.padovani@cs.unibo.it>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <assert.h>
#ifdef HAVE_IOSTREAM
#include <iostream>
#else
#include <stdio.h>
#endif

#include "Layout.hh"
#include "globals.hh"
#include "ValueConversion.hh"
#include "MathMLElement.hh"
#include "MathMLDocument.hh"
#include "MathMLAttributeList.hh"
#include "RenderingEnvironment.hh"
#include "MathMLStyleElement.hh"
#include "DrawingArea.hh"

#ifdef MATHML_STATS
unsigned MathMLElement::elementCounter = 0;
#endif

// MathMLElement: this is the base class for every MathML presentation element.
// It implements the basic skeleton of every such element, moreover it handles
// the attributes and provides some facility functions to access and parse
// attributes.
MathMLElement::MathMLElement(TagId t)
{
  tag = t;

  gc     = NULL; // the graphic context
  layout = NULL;
  shape  = NULL;

#ifdef MATHML_STATS
  elementCounter++;
#endif
}

MathMLElement::~MathMLElement()
{
  delete gc;
  delete layout;
  delete shape;

#ifdef MATHML_STATS
  elementCounter--;
#endif
}

// GetAttributeSignatureAux: this is an auxiliary function used to retrieve
// the signature of the attribute with id ID given an array of attribute
// signatures.
const AttributeSignature*
MathMLElement::GetAttributeSignatureAux(AttributeId id,
					AttributeSignature sig[]) const
{
  for (unsigned i = 0; sig[i].GetAttributeId() != ATTR_NOTVALID; i++)
    if (sig[i].GetAttributeId() == id) return &sig[i];

  return NULL;
}

// GetAttributeSignature: return the attribute signature of ID.
// Here are the attributes common to all (presentation) elements of MathML
const AttributeSignature*
MathMLElement::GetAttributeSignature(AttributeId id) const
{
  static AttributeSignature sig[] = {
    { ATTR_ID,         NULL, NULL },
    { ATTR_CLASS,      NULL, NULL },
    { ATTR_OTHER,      NULL, NULL },
#ifdef HELM_PROJECT
    { ATTR_HELM_HREF,  NULL, NULL },
#endif
    { ATTR_XLINK_XREF, NULL, NULL },

    { ATTR_NOTVALID, NULL, NULL }
  };

  return GetAttributeSignatureAux(id, sig);
}

void
MathMLElement::Normalize()
{
}

void
MathMLElement::Dump() const
{
  PrintIndentation();
#ifdef HAVE_IOSTREAM
  cout << '<' << NameOfTagId(tag);
#else
  printf("<%s", NameOfTagId(tag));
#endif
  attributes.Dump();
#ifdef HAVE_IOSTREAM
  cout << "/>" << endl;
#else
  printf("/>\n");
#endif
}

const Value*
MathMLElement::ParseAttribute(AttributeId id, const String* sValue) const
{
  assert(sValue != NULL);

  const AttributeSignature* aSignature = GetAttributeSignature(id);
  assert(aSignature != NULL);

  AttributeParser parser = aSignature->GetParser();
  assert(parser != NULL);

  StringTokenizer st(*sValue);
  const Value* value = parser(st);

  return value;
}

const String*
MathMLElement::GetAttribute(AttributeId id,
			    const RenderingEnvironment* env,
			    bool searchDefault) const
{
  const String* sValue = NULL;

  const AttributeSignature* aSignature = GetAttributeSignature(id);
  assert(aSignature != NULL);

  // first we have to search within local attributes, we do so
  const MathMLAttribute* attribute = attributes.GetAttribute(id);
  if (attribute != NULL) sValue = attribute->GetValue();

  if (sValue == NULL && env != NULL) {
    // attribute not found, see if someone set it in the environment
    const MathMLAttribute* attribute = env->GetAttribute(id);
    if (attribute != NULL) sValue = attribute->GetValue();
  }

  if (searchDefault) {
    if (sValue == NULL) {
      // still here, take the default value
      sValue = aSignature->GetDefaultValue();
    }

    assert(sValue != NULL);
  }

  return sValue;
}

const Value*
MathMLElement::GetDefaultAttributeValue(AttributeId id) const
{
  const AttributeSignature* aSignature = GetAttributeSignature(id);
  assert(aSignature != NULL);

  const String* sDefaultValue = aSignature->GetDefaultValue();
  assert(sDefaultValue != NULL);

  const Value* value = ParseAttribute(id, sDefaultValue);
  assert(value != NULL); // it is a default value, cannot be wrong

  return value;
}

const Value*
MathMLElement::GetAttributeValue(AttributeId id,
				 const RenderingEnvironment* env,
				 bool searchDefault) const
{
  const String* sValue = GetAttribute(id, env, searchDefault);

  // this is the only case when this function returns NULL
  if (!searchDefault && sValue == NULL) return NULL;

  assert(sValue != NULL);
  const Value* value = ParseAttribute(id, sValue);

  if (value == NULL) {
    if (verbose) {
#ifdef HAVE_IOSTREAM
      cerr << "WARNING: <" << NameOfTagId(IsA());
      cerr << ' ' << NameOfAttributeId(id) << " = ";
      cerr << "\"" << sValue->ToStaticC() << "\">, attribute parsing error (getting default)" << endl;
#else
      fprintf(stderr, "WARNING: <%s ", NameOfTagId(IsA()));
      fprintf(stderr, "%s = ", NameOfAttributeId(id));
      fprintf(stderr, "\"%s\">, attribute parsing error (getting default)\n", sValue->ToStaticC());
#endif
    }

    if (searchDefault) {
      // something went wrong while parsing, let's try with the default value
      value = GetDefaultAttributeValue(id);
      // the default value cannot be wrong
      assert(value != NULL);
    }
  }

  return value;
}

const Value*
MathMLElement::ResolveAttributeValue(AttributeId id,
				     const RenderingEnvironment* env) const
{
  assert(env != NULL);

  // ok, let's see what kind of attribute we have
  const Value* value = GetAttributeValue(id, env, true);
  assert(value != NULL);

  // if it is a number with unit and the unit is not %, then we are done
  if (value->IsNumberUnit() && !value->ToNumberUnit().IsPercentage()) return value;

  const Value* realValue = NULL;

  if (value->IsNumberUnit() && value->ToNumberUnit().IsPercentage()) {
    // it is a % of the default value
    const Value* defaultValue = GetDefaultAttributeValue(id);
    assert(defaultValue->IsNumberUnit());
    UnitValue defaultV = defaultValue->ToNumberUnit();
    UnitValue v = value->ToNumberUnit();
    realValue = new Value(defaultV.GetValue() * v.GetValue(),
			  defaultV.GetUnitId());
  }
  if (realValue == NULL) realValue = value;
  else delete value;

  return realValue;
}

const Value*
MathMLElement::Resolve(const Value* value,
		       const RenderingEnvironment* env,
		       int i, int j)
{
  assert(value != NULL);
  assert(env != NULL);

  const Value* realValue = value->Get(i, j);
  assert(realValue != NULL);

  if      (realValue->IsKeyword(KW_VERYVERYTHINMATHSPACE))
    realValue = new Value(env->GetMathSpace(MATH_SPACE_VERYVERYTHIN));
  else if (realValue->IsKeyword(KW_VERYTHINMATHSPACE))
    realValue = new Value(env->GetMathSpace(MATH_SPACE_VERYTHIN));
  else if (realValue->IsKeyword(KW_THINMATHSPACE))
    realValue = new Value(env->GetMathSpace(MATH_SPACE_THIN));
  else if (realValue->IsKeyword(KW_MEDIUMMATHSPACE))
    realValue = new Value(env->GetMathSpace(MATH_SPACE_MEDIUM));
  else if (realValue->IsKeyword(KW_THICKMATHSPACE))
    realValue = new Value(env->GetMathSpace(MATH_SPACE_THICK));
  else if (realValue->IsKeyword(KW_VERYTHICKMATHSPACE))
    realValue = new Value(env->GetMathSpace(MATH_SPACE_VERYTHICK));
  else if (realValue->IsKeyword(KW_VERYVERYTHICKMATHSPACE))
    realValue = new Value(env->GetMathSpace(MATH_SPACE_VERYVERYTHICK));
  else
    realValue = new Value(*realValue);

  return realValue;
}

bool
MathMLElement::IsSet(AttributeId id) const
{
  return attributes.GetAttribute(id) != NULL;
}

void
MathMLElement::Setup(RenderingEnvironment* env)
{
  // this function is defined to be empty but not pure-virtual
  // because some "space-like" elements such as <mspace>
  // <maligngroup>, <malignmark> effectively do nothing.
  // So we don't have to implement this function as empty
  // in every such element.
  // The same holds for Render below.
}

void
MathMLElement::ResetLayout()
{
  delete layout;
  layout = NULL;
}

void
MathMLElement::DoBoxedLayout(LayoutId id, BreakId bid, scaled maxWidth)
{
  ResetLayout();

  layout = new Layout(maxWidth, IsBreakable() ? bid : BREAK_NO);
  DoLayout(id, *layout);
  layout->DoLayout(id);
  if (id == LAYOUT_AUTO) DoStretchyLayout();
  layout->GetBoundingBox(box, id);

  ConfirmLayout(id);

#if 0
  cout << '`' << NameOfTagId(IsA()) << "' (" << this << ") DoBoxedLayout " << box << endl;
#endif
}

void
MathMLElement::DoLayout(LayoutId id, Layout& layout)
{
  // Well, there are some empty elements, such as <none/> or <prescripts/>
  // that do not have a layout, nonetheless they are unbreakable

  //cout << NameOfTagId(IsA()) << " do layout" << endl;
}

void
MathMLElement::RecalcBoundingBox(LayoutId id, scaled minWidth)
{
  assert(HasLayout());
  layout->GetBoundingBox(box, id);
  box.width = scaledMax(box.width, minWidth);
  ConfirmLayout(id);
}

void
MathMLElement::DoStretchyLayout()
{
}

void
MathMLElement::SetPosition(scaled x, scaled y)
{
  position.x = x;
  position.y = y;
  if (HasLayout()) layout->SetPosition(x, y);
}

void
MathMLElement::RenderBackground(const DrawingArea& area)
{
  assert(GetParent() != NULL);

  if (gc == NULL) {
    gc = area.CreateGraphicsContext();
    assert(gc != NULL);
  }

  if (HasDirtyBackground()) {
    assert(IsShaped());

    if (IsSelected()) gc->SetColor(SELECTED_COLOR);
    else gc->SetColor(GetParentBackgroundColor());

    if (IsSelected())
      cout << NameOfTagId(IsA()) << " rendering selected background" << endl;
    else
      cout << NameOfTagId(IsA()) << " rendering background" << endl;

    area.Clear(gc, GetShape());
  }
}

void
MathMLElement::Render(const DrawingArea& area)
{
  if (!IsDirty()) return;

  RenderBackground(area);

  ResetDirty();
}

void
MathMLElement::Freeze()
{
  assert(!IsBreakable() || HasLayout());

  if (shape != NULL) delete shape;
  
  Rectangle* rect = new Rectangle;
  GetBoundingBox().ToRectangle(GetX(), GetY(), *rect);
  shape = new Shape(rect);

#if 0
  cout << NameOfTagId(IsA()) << ' ' << sp2ipx(GetX()) << ',' << sp2ipx(GetY()) << M": freeze " << *shape << endl;
#endif

  ResetLayout();
}

void
MathMLElement::SetDirty(const Rectangle* rect)
{
  assert(IsShaped());

  dirtyBackground =
    (GetParent() != NULL && (GetParent()->IsSelected() != IsSelected())) ? 1 : 0;
#if 0
  if (GetParent() != NULL && (GetParent()->IsSelected() != IsSelected()))
    dirtyBackground = 1;
#endif

  if (IsDirty()) return;
  if (rect != NULL && !shape->Overlaps(*rect)) return;

  dirty = 1;
  SetDirtyChildren();
}

bool
MathMLElement::IsSpaceLike() const
{
  return false;
}

bool
MathMLElement::IsElement() const
{
  return true;
}

bool
MathMLElement::IsExpanding() const
{
  return false;
}

bool
MathMLElement::IsInside(scaled x, scaled y) const
{
  assert(IsShaped());
  return shape->IsInside(x, y);
}

void
MathMLElement::GetLinearBoundingBox(BoundingBox& b) const
{
  assert(!IsBreakable());
  b = box;
}

MathMLElement*
MathMLElement::Inside(scaled x, scaled y)
{
  return IsInside(x, y) ? this : NULL;
}

unsigned
MathMLElement::GetDepth() const
{
  unsigned depth = 0;
  const MathMLElement* p = this;
  
  while (p != NULL) {
    depth++;
    p = p->GetParent();
  }

  return depth;
}

RGBValue
MathMLElement::GetParentBackgroundColor() const
{
  const MathMLElement* p = GetParent();

  while (p != NULL && p->IsA() != TAG_MSTYLE) p = p->GetParent();

  if (p == NULL) return WHITE_COLOR;

  const MathMLStyleElement* elem = dynamic_cast<const MathMLStyleElement*>(p);
  assert(elem != NULL);

  return elem->GetBackgroundColor();
}

const Layout&
MathMLElement::GetLayout() const
{
  assert(HasLayout());
  return *layout;
}

const Shape&
MathMLElement::GetShape() const
{
  assert(IsShaped());
  return *shape;
}

scaled
MathMLElement::GetLeftEdge() const
{
  return GetX();
}

scaled
MathMLElement::GetRightEdge() const
{
  return GetX();
}
