// Copyright (C) 2000, Luca Padovani <luca.padovani@cs.unibo.it>.
// 
// This file is part of GtkMathView, a Gtk widget for MathML.
// 
// GtkMathView is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// GtkMathView is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with GtkMathView; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// 
// For details, see the GtkMathView World-Wide-Web page,
// http://cs.unibo.it/~lpadovan/mml-widget, or send a mail to
// <luca.padovani@cs.unibo.it>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <assert.h>
#ifdef HAVE_IOSTREAM
#include <iostream>
#include <iomanip.h>
#else
#include <stdio.h>
#endif

#include "globals.hh"
#include "Iterator.hh"
#include "operatorAux.hh"
#include "ValueSequence.hh"
#include "StringUnicode.hh"
#include "MathMLCharNode.hh"
#include "ValueConversion.hh"
#include "MathMLStringNode.hh"
#include "MathMLRowElement.hh"
#include "OperatorDictionary.hh"
#include "RenderingEnvironment.hh"
#include "MathMLOperatorElement.hh"
#include "MathMLStretchyCharNode.hh"
#include "MathMLEmbellishedOperatorElement.hh"

MathMLOperatorElement::MathMLOperatorElement() : MathMLTokenElement(TAG_MO)
{
  eOp = NULL;
  defaults = NULL;
  operatorName = NULL;
}

MathMLOperatorElement::~MathMLOperatorElement()
{
  delete operatorName;
}

const AttributeSignature*
MathMLOperatorElement::GetAttributeSignature(AttributeId id) const
{
  static AttributeSignature sig[] = {
    { ATTR_FORM,      	  operatorFormParser, 	 NULL },
    { ATTR_FENCE,     	  booleanParser,      	 new StringC("false") },
    { ATTR_SEPARATOR, 	  booleanParser,      	 new StringC("false") },
    { ATTR_LSPACE,    	  spaceParser,    	 new StringC("thickmathspace") },
    { ATTR_RSPACE,    	  spaceParser,    	 new StringC("thickmathspace") },
    { ATTR_STRETCHY,  	  booleanParser,      	 new StringC("false") },
    { ATTR_SYMMETRIC, 	  booleanParser,      	 new StringC("true") },
    { ATTR_MAXSIZE,   	  operatorMaxSizeParser, new StringC("infinity") },
    { ATTR_MINSIZE,   	  operatorMinSizeParser, new StringC("1") },
    { ATTR_LARGEOP,       booleanParser,         new StringC("false") },
    { ATTR_MOVABLELIMITS, booleanParser,         new StringC("false") },
    { ATTR_ACCENT,        booleanParser,         new StringC("false") },
    { ATTR_NOTVALID,      NULL,                  NULL }
  };

  const AttributeSignature* signature = GetAttributeSignatureAux(id, sig);
  if (signature == NULL) signature = MathMLTokenElement::GetAttributeSignature(id);

  return signature;
}

void
MathMLOperatorElement::Normalize()
{
  MathMLElement* root = getEmbellishedOperatorRoot(this);
  assert(root != NULL);

  MathMLElement* p = root->GetParent();
  assert(p != NULL);

  MathMLEmbellishedOperatorElement* op = new MathMLEmbellishedOperatorElement(this);
  op->SetParent(p);
  op->content.Append(root);
  eOp = op;

  // now we have to substitute the root of the embellished operator
  // with the embellished operator object just created
  assert(p->IsContainer());
  MathMLContainerElement* pContainer = TO_CONTAINER(p);
  assert(pContainer != NULL);
  pContainer->content.Substitute(root, eOp);
  // Important!!! The root of the embellished operator now has
  // the "embellishment" as parent
  root->SetParent(eOp);

  MathMLTokenElement::Normalize();
}

void
MathMLOperatorElement::Setup(RenderingEnvironment* env)
{
  assert(env != NULL);

  const Value* value = NULL;
  const Value* resValue = NULL;

  axis = env->GetAxis();

  value = GetAttributeValue(ATTR_FORM, env, false);
  if (value != NULL) form = ToFormId(value);
  else form = InferOperatorForm();
  delete value;

  operatorName = FlattenContent();

  const OperatorDefaults* prefix  = NULL;
  const OperatorDefaults* infix   = NULL;
  const OperatorDefaults* postfix = NULL;
  
  if (operatorName != NULL)
    dictionary.Search(operatorName, &prefix, &infix, &postfix);

  if      (form == OP_FORM_PREFIX && prefix != NULL) defaults = prefix;
  else if (form == OP_FORM_INFIX && infix != NULL) defaults = infix;
  else if (form == OP_FORM_POSTFIX && postfix != NULL) defaults = postfix;
  else if (infix != NULL) defaults = infix;
  else if (postfix != NULL) defaults = postfix;
  else if (prefix != NULL) defaults = prefix;
  else defaults = NULL;

  value = GetOperatorAttributeValue(ATTR_FENCE, env);
  assert(value != NULL && value->IsBoolean());
  fence = value->ToBoolean();
  delete value;

  value = GetOperatorAttributeValue(ATTR_SEPARATOR, env);
  assert(value != NULL && value->IsBoolean());
  separator = value->ToBoolean();
  delete value;

  value = GetOperatorAttributeValue(ATTR_LSPACE, env);
  assert(value != NULL);
  resValue = Resolve(value, env);
  assert(resValue != NULL && resValue->IsNumberUnit());
  lSpace = env->ToScaledPoints(resValue->ToNumberUnit());
  delete value;
  delete resValue;

  value = GetOperatorAttributeValue(ATTR_RSPACE, env);
  assert(value != NULL);
  resValue = Resolve(value, env);
  assert(resValue != NULL && resValue->IsNumberUnit());
  rSpace = env->ToScaledPoints(resValue->ToNumberUnit());
  delete value;
  delete resValue;

  value = GetOperatorAttributeValue(ATTR_STRETCHY, env);
  assert(value != NULL && value->IsBoolean());
  stretchy = value->ToBoolean();
  delete value;

  value = GetOperatorAttributeValue(ATTR_SYMMETRIC, env);
  assert(value != NULL && value->IsBoolean());
  symmetric = value->ToBoolean();
  delete value;

  value = GetOperatorAttributeValue(ATTR_MAXSIZE, env);
  assert(value != NULL);
  if (value->IsKeyword(KW_INFINITY)) infiniteMaxSize = true;
  else {
    infiniteMaxSize = false;
    ParseLimitValue(value, env, maxMultiplier, maxSize);
  }
  delete value;

  value = GetOperatorAttributeValue(ATTR_MINSIZE, env);
  assert(value != NULL);
  ParseLimitValue(value, env, minMultiplier, minSize);

  value = GetOperatorAttributeValue(ATTR_MOVABLELIMITS, env);
  assert(value != NULL && value->IsBoolean());
  movableLimits = value->ToBoolean();
  delete value;

  value = GetOperatorAttributeValue(ATTR_ACCENT, env);
  assert(value != NULL && value->IsBoolean());
  accent = value->ToBoolean();  
  delete value;

  value = GetOperatorAttributeValue(ATTR_LARGEOP, env);
  assert(value != NULL && value->IsBoolean());
  bool largeOp = value->ToBoolean();
  delete value;

  MathMLTokenElement::Setup(env);

  if (content.GetSize() == 1 && largeOp && env->GetDisplayStyle()) {
    assert(content.GetFirst() != NULL);
    if (content.GetFirst()->IsStretchyChar()) {
      MathMLStretchyCharNode* sNode = TO_STRETCHY_CHAR(content.GetFirst());
      assert(sNode != NULL);
      sNode->SetDefaultLargeGlyph(true);
    }
  }
}

void
MathMLOperatorElement::VerticalStretchTo(scaled ascent, scaled descent, bool strict)
{
  assert(stretchy);

  scaled desiredSize = 0;

  // ok, if we have to stretch to a very tiny symbol, then we do not
  // stretch at all. This simplifies subsequent computations
  if (ascent + descent < axis) return;

  // Here we have to calculate the desired size of the stretchable symbol.
  // If symmetric == true the we have to stretch to cover the maximum among
  // height and depth, otherwise we just stretch to ascent + descent
  desiredSize = symmetric ? (2 * scaledMax(ascent - axis, descent + axis)) : (ascent + descent);

  desiredSize = scaledMax(SP_EPSILON, desiredSize);

  // ...however, there may be some contraints over the size of the stretchable
  // operator. adjustedSize will be the final allowed size for the operator
  scaled minHeight = GetMinBoundingBox().GetHeight();
  scaled adjustedSize = desiredSize;

  if (minMultiplier > 0)
    adjustedSize = scaledMax(adjustedSize, float2sp(minHeight * minMultiplier));
  else
    adjustedSize = scaledMax(adjustedSize, minSize);

  if (!infiniteMaxSize) {
    if (maxMultiplier > 0)
      adjustedSize = scaledMin(adjustedSize, float2sp(minHeight * maxMultiplier));
    else
      adjustedSize = scaledMin(adjustedSize, maxSize);
  }

  adjustedSize = scaledMax(0, adjustedSize);

  assert(content.GetSize() == 1);
  assert(content.GetFirst() != NULL);
  if (!content.GetFirst()->IsStretchyChar()) {
    if (verbose) {
#ifdef HAVE_IOSTREAM
      cerr << "WARNING: character ";
      cerr << "`U+" << setw(4) << setfill('0') << hex << operatorName->GetChar(0) << dec;
      cerr << "' could not be stretched" << endl;
#else
      fprintf(stderr, "WARNING: character `U+%04x' could not be stretched\n", operatorName->GetChar(0));
#endif
    }
    return;
  }

  MathMLStretchyCharNode* sNode = TO_STRETCHY_CHAR(content.GetFirst());
  assert(sNode != NULL);

  float ascentR = 1.0;
  if (symmetric)
    ascentR = sp2float(scaledMax(ascent - axis, descent + axis)) / sp2float(desiredSize);
  else
    ascentR = sp2float(scaledMax(0, ascent - axis)) / sp2float(desiredSize);

  sNode->DoVerticalStretchyLayout(adjustedSize, ascentR, axis, strict);
}

void
MathMLOperatorElement::HorizontalStretchTo(scaled width, bool strict)
{
  assert(stretchy);

  scaled desiredSize = width;

  desiredSize = scaledMax(SP_EPSILON, desiredSize);

  // ...however, there may be some contraints over the size of the stretchable
  // operator. adjustedSize will be the final allowed size for the operator
  scaled minWidth = GetMinBoundingBox().width;
  scaled adjustedSize = desiredSize;

  if (minMultiplier > 0)
    adjustedSize = scaledMax(adjustedSize, float2sp(minWidth * minMultiplier));
  else
    adjustedSize = scaledMax(adjustedSize, minSize);

  if (!infiniteMaxSize) {
    if (maxMultiplier > 0)
      adjustedSize = scaledMin(adjustedSize, float2sp(minWidth * maxMultiplier));
    else
      adjustedSize = scaledMin(adjustedSize, minSize);
  }

  adjustedSize = scaledMax(0, adjustedSize);

  assert(content.GetSize() == 1);
  assert(content.GetFirst() != NULL);
  if (!content.GetFirst()->IsStretchyChar()) {
    if (verbose) {
#ifdef HAVE_IOSTREAM
      cerr << "WARNING: character ";
      cerr << "`U+" << setw(4) << setfill('0') << hex << operatorName->GetChar(0) << dec;
      cerr << "' could not be stretched" << endl;
#else
      fprintf(stderr, "WARNING: character `U+%04x' could not be stretched\n", operatorName->GetChar(0));
#endif
    }
    return;
  }

  MathMLStretchyCharNode* sNode = TO_STRETCHY_CHAR(content.GetFirst());
  assert(sNode != NULL);

  // now we do the stretchy layout. fontAttributes will be used to find the
  // proper font
  sNode->DoHorizontalStretchyLayout(adjustedSize, strict);
}

void
MathMLOperatorElement::ParseLimitValue(const Value* value,
				       const RenderingEnvironment* env,
				       float& multiplier,
				       scaled& size)
{
  assert(value != NULL);
  assert(env != NULL);

  if (value->IsKeyword()) { // it must be a named math space
    const Value* resValue = Resolve(value, env);
    assert(resValue->IsNumberUnit());
    multiplier = -1;
    size = env->ToScaledPoints(resValue->ToNumberUnit());
    delete resValue;
  } else {
    assert(value->IsSequence());
    const ValueSequence* seq = value->ToValueSequence();
    const Value* v = seq->GetFirstValue();
    const Value* unitV = seq->GetLastValue();
    assert(v != NULL);
    assert(v->IsNumber());
    assert(unitV != NULL);
    
    if (unitV->IsEmpty()) {      
      multiplier = floatMax(EPSILON, v->ToNumber());
    } else {
      assert(unitV->IsKeyword());
      multiplier = -1;

      UnitValue siz;
      siz.Set(v->ToNumber(), ToUnitId(unitV));

      if (siz.IsPercentage()) {
#ifdef HAVE_IOSTREAM
	cout << "WARNING: percentage value specified in maxsize attribute (mo), ignored" << endl;
#else
	printf("WARNING: percentage value specified in maxsize attribute (mo), ignored\n");
#endif
	multiplier = floatMax(EPSILON, siz.GetValue());
      } else {
	multiplier = -1;
	size = env->ToScaledPoints(siz);
      }
    }
  }
}

const Value*
MathMLOperatorElement::GetOperatorAttributeValue(AttributeId id,
						 const RenderingEnvironment* env) const
{
  assert(env != NULL);

  const String* sValue = NULL;

  sValue = GetAttribute(id, NULL, false);

  if (defaults != NULL && sValue == NULL) {
    switch (id) {
    case ATTR_FENCE: sValue = defaults->fence; break;
    case ATTR_SEPARATOR: sValue = defaults->separator; break;
    case ATTR_LSPACE: sValue = defaults->lspace; break;
    case ATTR_RSPACE: sValue = defaults->rspace; break;
    case ATTR_STRETCHY: sValue = defaults->stretchy; break;
    case ATTR_SYMMETRIC: sValue = defaults->symmetric; break;
    case ATTR_MAXSIZE: sValue = defaults->maxsize; break;
    case ATTR_MINSIZE: sValue = defaults->minsize; break;
    case ATTR_LARGEOP: sValue = defaults->largeop; break;
    case ATTR_MOVABLELIMITS: sValue = defaults->movablelimits; break;
    case ATTR_ACCENT: sValue = defaults->accent; break;
    default: assert(IMPOSSIBLE); break;
    }
  }

  if (sValue == NULL) sValue = GetAttribute(id, env);
  assert(sValue != NULL);

  const Value* value = ParseAttribute(id, sValue);
  assert(value != NULL);

  return value;
}

OperatorFormId
MathMLOperatorElement::InferOperatorForm() const
{
  assert(eOp != NULL);

  MathMLElement* elem = eOp->GetParent();
  assert(elem != NULL);

  OperatorFormId res = OP_FORM_INFIX;

  if (elem->IsA() == TAG_MROW) {
    MathMLRowElement* row = TO_ROW(elem);
    assert(row != NULL);

    unsigned rowLength = 0;
    unsigned position  = 0;
    for (Iterator<MathMLElement*> i(row->content); i.More(); i.Next()) {
      const MathMLElement* p = i();
      assert(p != NULL);

      if (!p->IsSpaceLike()) {
	if (p == eOp) position = rowLength;
	rowLength++;
      }
    }
    
    if (rowLength > 1) {
      if (position == 0) res = OP_FORM_PREFIX;
      else if (position == rowLength - 1) res = OP_FORM_POSTFIX;
    }
  }

  return res;
}

const String*
MathMLOperatorElement::FlattenContent() const
{
  const String* s = NULL;

  for (Iterator<MathMLFrame*> i(content); i.More(); i.Next()) {
    MathMLFrame* node = i();
    assert(node != NULL);

    if (s == NULL) {
      if (node->IsString()) {
	MathMLStringNode* text = TO_STRING(node);
	assert(text != NULL);

	const String* textS = text->GetContent();
	assert(textS != NULL);
	s = textS->Clone();
      } else if (node->IsChar()) {
	MathMLCharNode* text = TO_CHAR(node);
	assert(text != NULL);

	Char ch = text->GetChar();
	s = new StringUnicode(&ch, 1);
      }
    } else {
#ifdef HAVE_IOSTREAM
      cout << "WARNING: multi chunk operator (not yet implemented)" << endl;
#else
      printf("WARNING: multi chunk operator (not yet implemented)\n");
#endif
      assert(NOT_IMPLEMENTED);
    }    
  }

  return s;
}

StretchId
MathMLOperatorElement::GetStretch() const
{
  if (!IsStretchy()) return STRETCH_NO;

  assert(content.GetSize() == 1);
  assert(content.GetFirst() != NULL);

  if (!content.GetFirst()->IsStretchyChar()) return STRETCH_NO;

  MathMLStretchyCharNode* sChar = TO_STRETCHY_CHAR(content.GetFirst());
  assert(sChar != NULL);

  return sChar->GetStretch();
}

bool
MathMLOperatorElement::IsOperator() const
{
  return true;
}

bool
MathMLOperatorElement::IsBreakable() const
{
  return false;
}

