// Copyright (C) 2000, Luca Padovani <luca.padovani@cs.unibo.it>.
// 
// This file is part of GtkMathView, a Gtk widget for MathML.
// 
// GtkMathView is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// GtkMathView is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with GtkMathView; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// 
// For details, see the GtkMathView World-Wide-Web page,
// http://cs.unibo.it/~lpadovan/mml-widget, or send a mail to
// <luca.padovani@cs.unibo.it>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <assert.h>
#include <stddef.h>

#include "MathMLRowElement.hh"
#include "MathMLRadicalElement.hh"
#include "RenderingEnvironment.hh"

MathMLRadicalElement::MathMLRadicalElement(TagId id) : MathMLContainerElement(id)
{
  assert(id == TAG_MSQRT || id == TAG_MROOT);
}

MathMLRadicalElement::~MathMLRadicalElement()
{
}

void
MathMLRadicalElement::Normalize()
{
  unsigned n = (IsA() == TAG_MSQRT) ? 1 : 2;

  while (content.GetSize() < n) {
    MathMLRowElement* mrow = new MathMLRowElement;
    mrow->SetParent(this);
    content.Append(mrow);
  }

  while (content.GetSize() > n) {
    MathMLElement* elem = content.RemoveLast();
    delete elem;
  }

  MathMLContainerElement::Normalize();
}

void
MathMLRadicalElement::Setup(RenderingEnvironment* env)
{
  assert(env != NULL);

  sppm    = env->GetScaledPointsPerEm();
  spacing = env->ToScaledPoints(env->GetMathSpace(MATH_SPACE_MEDIUM));
  color   = env->GetColor();

  MathMLElement* base = content.GetFirst();
  assert(base != NULL);

  base->Setup(env);

  if (IsA() == TAG_MROOT) {
    MathMLElement* script = content.GetLast();
    assert(script != NULL);
    
    env->Push();
    env->SetDisplayStyle(false);
    env->AddScriptLevel(2);
    script->Setup(env);
    env->Drop();
  }
}

void
MathMLRadicalElement::DoBoxedLayout(LayoutId id, BreakId bid, scaled maxWidth)
{
  scaled available = scaledMax(0, maxWidth - 2 * spacing - sppm / 2);

  MathMLElement* base = content.GetFirst();
  assert(base != NULL);

  base->DoBoxedLayout(id, BREAK_NO, available / 2);
  const BoundingBox& baseBox = base->GetBoundingBox();

  if (IsA() == TAG_MROOT) {
    MathMLElement* script = content.GetLast();
    assert(script != NULL);

    script->DoBoxedLayout(id, BREAK_NO, available / 2);
    const BoundingBox& scriptBox = script->GetBoundingBox();

    DoBaseLayoutAux(box, baseBox);

    if (scriptBox.width > sppm / 2)
      box.width += scriptBox.width - sppm / 2 - spacing;

    if (baseBox.GetHeight() / 2 < scriptBox.GetHeight() + spacing) {
      box.ascent += scriptBox.GetHeight() + spacing - baseBox.GetHeight() / 2;
    }
  } else {
    DoBaseLayoutAux(box, baseBox);
  }

  ConfirmLayout(id);
}

void
MathMLRadicalElement::DoBaseLayoutAux(BoundingBox& b, const BoundingBox& baseBox)
{
  b.Zero();
  b.ascent  = baseBox.ascent + spacing;
  b.descent = baseBox.descent;
  b.width   = sppm + baseBox.width + 2 * spacing;
}

void
MathMLRadicalElement::SetPosition(scaled x, scaled y)
{
  position.x = x;
  position.y = y;

  MathMLElement* base = content.GetFirst();
  assert(base != NULL);

  if (IsA() == TAG_MROOT) {
    MathMLElement* script = content.GetLast();
    assert(script != NULL);

    const BoundingBox& baseBox   = base->GetBoundingBox();
    const BoundingBox& scriptBox = script->GetBoundingBox();

    if (scriptBox.width > sppm / 2) {
      script->SetPosition(x,
			  y + (baseBox.GetHeight() / 2 - baseBox.ascent) - scriptBox.descent - 2 * spacing);
      base->SetPosition(x + scriptBox.width + sppm / 2, y);
    } else {
      script->SetPosition(x + sppm / 2 - scriptBox.width,
			  y + (baseBox.GetHeight() / 2 - baseBox.ascent) - scriptBox.descent - 2 * spacing);
      base->SetPosition(x + sppm + spacing, y);
    }
  } else {
    base->SetPosition(x + sppm + spacing, y);
  }
}

void
MathMLRadicalElement::Render(const DrawingArea& area)
{
  if (!HasDirtyChildren()) return;

  MathMLContainerElement::Render(area);
  assert(gc != NULL);

  MathMLElement* base = content.GetFirst();
  assert(base != NULL);

  const BoundingBox& baseBox = base->GetBoundingBox();

  gc->SetColor(color);

  scaled x = position.x;
  scaled y = position.y;
  scaled xOffset = 0;

  if (IsA() == TAG_MROOT) {
    MathMLElement* script = content.GetLast();
    assert(script != NULL);

    const BoundingBox& scriptBox = script->GetBoundingBox();

    if (scriptBox.width > sppm / 2)
      xOffset = scriptBox.width - sppm / 2 - spacing;
  }

  assert(scaledGeq(xOffset, 0));

  area.MoveTo(x + xOffset, y + baseBox.GetHeight() / 2 - baseBox.ascent);
  area.DrawLineToDelta(gc, spacing, -spacing);
  gc->SetLineStyle(LINE_STYLE_SOLID, 2);
  area.DrawLineTo(gc, x + xOffset + sppm / 2, y + box.descent);
  gc->SetLineStyle(LINE_STYLE_SOLID, 1);
  area.DrawLineTo(gc, x + xOffset + sppm, y - baseBox.ascent - spacing);
  area.DrawLineTo(gc, x + xOffset + sppm + 2 * spacing + baseBox.width, y - baseBox.ascent - spacing);
}

bool
MathMLRadicalElement::IsExpanding() const
{
  MathMLElement* base = content.GetFirst();
  assert(base != NULL);
  return base->IsExpanding();
}

scaled
MathMLRadicalElement::GetLeftEdge() const
{
  return scaledMin(MathMLContainerElement::GetLeftEdge(), GetX());
}

BreakId
MathMLRadicalElement::GetBreakability() const
{
  return BREAK_AUTO;
}
