// Copyright (C) 2000, Luca Padovani <luca.padovani@cs.unibo.it>.
// 
// This file is part of GtkMathView, a Gtk widget for MathML.
// 
// GtkMathView is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// GtkMathView is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with GtkMathView; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// 
// For details, see the GtkMathView World-Wide-Web page,
// http://cs.unibo.it/~lpadovan/mml-widget, or send a mail to
// <luca.padovani@cs.unibo.it>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_IOSTREAM
#include <iostream>
#else
#include <stdio.h>
#endif

#include <assert.h>

#include "globals.hh"
#include "Iterator.hh"
#include "MathMLRowElement.hh"
#include "MathMLTableRowElement.hh"
#include "MathMLTableCellElement.hh"

MathMLTableRowElement::MathMLTableRowElement(TagId id): MathMLContainerElement(id)
{
  assert(id == TAG_MTR || id == TAG_MLABELEDTR);
  rowIndex = 0;
}

MathMLTableRowElement::~MathMLTableRowElement()
{
}

const AttributeSignature*
MathMLTableRowElement::GetAttributeSignature(AttributeId id) const
{
  static AttributeSignature sig[] = {
    { ATTR_ROWALIGN,        rowAlignListParser,       NULL },
    { ATTR_COLUMNALIGN,     columnAlignListParser,    NULL },
    { ATTR_GROUPALIGN,      groupAlignListListParser, NULL },
    { ATTR_NOTVALID,        NULL,                     NULL }
  };

  const AttributeSignature* signature = GetAttributeSignatureAux(id, sig);
  if (signature == NULL) signature = MathMLElement::GetAttributeSignature(id);

  return signature;
}

void
MathMLTableRowElement::Normalize()
{
  if (IsA() == TAG_MLABELEDTR &&
      (content.GetSize() == 0 ||
       (content.GetSize() > 0 &&
	content.GetFirst() != NULL &&
	content.GetFirst()->IsA() == TAG_MTD))) {
    if (verbose) {
#ifdef HAVE_IOSTREAM
      cerr << "WARNING: `mlabeledtr' element without label (empty label added)" << endl;
#else
      fprintf(stderr, "WARNING: `mlabeledtr' element without label (empty label added)\n");
#endif
    }

    MathMLRowElement* mrow = new MathMLRowElement();
    mrow->SetParent(this);
    content.AddFirst(mrow);
  }

  for (unsigned i = 0; i < content.GetSize(); i++) {
    MathMLElement* elem = content.RemoveFirst();
    assert(elem != NULL);

    // if this is a labeled row, then the first child is always the label
    // because of normalization (see above)
    if (elem->IsA() != TAG_MTD && (IsA() == TAG_MTR || i > 0)) {
      MathMLTableCellElement* inferredTableCell = new MathMLTableCellElement;
      inferredTableCell->SetParent(this);
      inferredTableCell->content.Append(elem);

      elem->SetParent(inferredTableCell);
      elem = inferredTableCell;
    }
    elem->Normalize();
    content.Append(elem);
  }
}

void
MathMLTableRowElement::SetupRowIndex(unsigned i)
{
  rowIndex = i;
}

void
MathMLTableRowElement::SetupCellSpanning(RenderingEnvironment* env)
{
  for (Iterator<MathMLElement*> p(content); p.More(); p.Next()) {
    assert(p() != NULL);

    if (IsA() == TAG_MTR || p() != content.GetFirst()) {
      assert(p()->IsA() == TAG_MTD);

      MathMLTableCellElement* mtd = TO_TABLECELL(p());
      assert(mtd != NULL);

      mtd->SetupCellSpanning(env);
    }
  }
}

void
MathMLTableRowElement::Setup(RenderingEnvironment* env)
{
  assert(GetParent() != NULL);
  MathMLTableElement* mtable = TO_TABLE(GetParent());
  assert(mtable != NULL);

  const Value* value;

  value = GetAttributeValue(ATTR_COLUMNALIGN, NULL, false);
  if (value != NULL) mtable->SetupColumnAlignAux(value, rowIndex, 1, IsA() == TAG_MLABELEDTR);
  delete value;

  value = GetAttributeValue(ATTR_ROWALIGN, NULL, false);
  if (value != NULL) mtable->SetupRowAlignAux(value, rowIndex, 1, IsA() == TAG_MLABELEDTR);
  delete value;

  value = GetAttributeValue(ATTR_GROUPALIGN, NULL, false);
  if (value != NULL) mtable->SetupGroupAlignAux(value, rowIndex, 1);
  delete value;

  MathMLContainerElement::Setup(env);
}

