// Copyright (C) 2000, Luca Padovani <luca.padovani@cs.unibo.it>.
// 
// This file is part of GtkMathView, a Gtk widget for MathML.
// 
// GtkMathView is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// GtkMathView is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with GtkMathView; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// 
// For details, see the GtkMathView World-Wide-Web page,
// http://cs.unibo.it/~lpadovan/mml-widget, or send a mail to
// <luca.padovani@cs.unibo.it>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <assert.h>
#ifdef HAVE_IOSTREAM
#include <iostream>
#else
#include <stdio.h>
#endif

#include "Font.hh"
#include "Layout.hh"
#include "frameAux.hh"
#include "Iterator.hh"
#include "ShapeFactory.hh"
#include "StringUnicode.hh"
#include "MathMLTextNode.hh"
#include "ValueConversion.hh"
#include "MathMLStringNode.hh"
#include "MathMLTokenElement.hh"
#include "RenderingEnvironment.hh"
#include "MathMLStretchyCharNode.hh"

#ifdef MATHML_STATS
unsigned MathMLTokenElement::tokenCounter = 0;
#endif

MathMLTokenElement::MathMLTokenElement(TagId t) : MathMLElement(t)
{
  font = NULL;

#ifdef MATHML_STATS
  tokenCounter++;
#endif
}

MathMLTokenElement::~MathMLTokenElement()
{
  for (Iterator<MathMLFrame*> i(content); i.More(); i.Next()) delete i();
}

void
MathMLTokenElement::Dump() const
{
  PrintIndentation();
#ifdef HAVE_IOSTREAM
  cout << '<' << NameOfTagId(IsA());
#else
  printf("<%s", NameOfTagId(IsA()));
#endif
  attributes.Dump();
#ifdef HAVE_IOSTREAM
  cout << '>' << endl;
#else
  printf(">\n");
#endif

  IndentMore();
  for (Iterator<MathMLFrame*> i(content); i.More(); i.Next()) {
    assert(i() != NULL);
    if (i()->IsText()) {
      MathMLTextNode* text = TO_TEXT(i());
      assert(text != NULL);
      text->Dump();
    } else if (i()->IsElement()) {
      MathMLElement* elem = TO_ELEMENT(i());
      assert(elem != NULL);
      elem->Dump();
    }
  }
  IndentLess();

  PrintIndentation();
#ifdef HAVE_IOSTREAM
  cout << "</" << NameOfTagId(IsA()) << '>' << endl;
#else
  printf("</%s>\n", NameOfTagId(IsA()));  
#endif
}

const AttributeSignature*
MathMLTokenElement::GetAttributeSignature(AttributeId id) const
{
  static AttributeSignature sig[] = {
    { ATTR_FONTSIZE,   numberUnitParser, NULL },
    { ATTR_FONTWEIGHT, fontWeightParser, NULL },
    { ATTR_FONTSTYLE,  fontStyleParser,  new StringC("normal") },
    { ATTR_FONTFAMILY, stringParser,     NULL },
    { ATTR_COLOR,      colorParser,      NULL },
    { ATTR_NOTVALID,   NULL,             NULL }
  };

  const AttributeSignature* signature = GetAttributeSignatureAux(id, sig);
  if (signature == NULL) signature = MathMLElement::GetAttributeSignature(id);

  return signature;
}

unsigned
MathMLTokenElement::RawContentLength() const
{
  unsigned length = 0;

  for (Iterator<MathMLFrame*> i(content); i.More(); i.Next()) {
    MathMLFrame* node = i();
    assert(node != NULL);
    if (node->IsString()) {
      MathMLStringNode* text = TO_STRING(node);
      const String* sText = text->GetContent();
      assert(sText != NULL);
      length += sText->GetLength();
    } else if (node->IsChar()) {
      length++;
    }
  }

  return length;
}

void
MathMLTokenElement::Setup(RenderingEnvironment* env)
{
  env->Push();

  const Value* value = NULL;

  value = GetAttributeValue(ATTR_FONTSIZE, NULL, false);
  if (value != NULL) env->SetFontSize(value->ToNumberUnit());
  delete value;

  value = GetAttributeValue(ATTR_FONTFAMILY, NULL, false);
  if (value != NULL) env->SetFontFamily(value->ToString());
  delete value;

  value = GetAttributeValue(ATTR_FONTWEIGHT, NULL, false);
  if (value != NULL) env->SetFontWeight(ToFontWeightId(value));
  delete value;

  value = GetAttributeValue(ATTR_FONTSTYLE, NULL, false);
  if (value != NULL) env->SetFontStyle(ToFontStyleId(value));
  else if (IsA() == TAG_MI) {
    if (RawContentLength() == 1) env->SetFontStyle(FONT_STYLE_ITALIC);
    else env->SetFontStyle(FONT_STYLE_NORMAL);
  }
  delete value;

  value = GetAttributeValue(ATTR_COLOR, NULL, false);
  if (value != NULL) env->SetColor(ToRGB(value));
#ifdef HELM_PROJECT
  else if (IsSet(ATTR_HELM_HREF)) env->SetColor(LINK_COLOR);
#endif
  delete value;

  font = env->GetFont();
  assert(font != NULL);
  color = env->GetColor();

  const FontAttributes& fa = env->GetFontAttributes();
  for (Iterator<MathMLFrame*> i(content); i.More(); i.Next()) {
    assert(i() != NULL);
    if (i()->IsText()) {
      MathMLTextNode* text = TO_TEXT(i());
      assert(text != NULL);
      text->Fontify(fa);
    } else {
      // this is necessary since <mailgnmark/> elements can be inserted
      // inside token elements
      MathMLElement* elem = TO_ELEMENT(i());
      assert(elem != NULL);
      elem->Setup(env);
    }
  }

  env->Drop();
}

void
MathMLTokenElement::DoLayout(LayoutId id, Layout& layout)
{
  assert(font != NULL);

  scaled sppm = font->GetEm();
  scaled oldLineSpacing = layout.GetLineSpacing();
  layout.SetMinimumLineSpacing(font->GetSize());

  Iterator<MathMLFrame*> i(content);
  while (i.More()) {
    MathMLFrame* node = i();
    assert(node != NULL);
    if (node->IsText()) {
      MathMLTextNode* text = TO_TEXT(node);
      assert(text != NULL);

      scaled spacing = (sppm * text->GetSpacing()) / 18;
      BreakId breakability = text->GetBreakability();

      // ok, we do the actual layout of the chunk only if we are
      // doing minimum layout or if the chunk is indeed a stretchable
      // char. In all the other cases the layout is exactly the same
      // as the minimum layout, that we have previously done,
      // so we save some work.
      if (id == LAYOUT_MIN) {
	if (text->IsStretchyChar()) {
	  MathMLStretchyCharNode* node = TO_STRETCHY_CHAR(text);
	  assert(node != NULL);
	  node->DoMinimumLayout();
	} else
	  text->DoLayout();
      }

      if (!text->IsNonMarking()) layout.Append(text, breakability);
      if (text->GetSpacing() != 0) layout.Append(spacing, breakability);
      else layout.SetLastBreakability(breakability);
    }

    i.Next();
  }

  layout.SetLineSpacing(oldLineSpacing);
}

void
MathMLTokenElement::Freeze()
{
  if (HasLayout()) MathMLElement::Freeze();
  else {
    if (shape != NULL) delete shape;
    ShapeFactory shapeFactory;
    for (Iterator<MathMLFrame*> i(content); i.More(); i.Next()) {
      assert(i() != NULL);

      if (i()->IsElement()) {
	// malignmark is the only element allowed inside a token
	MathMLElement* elem = TO_ELEMENT(i());
	assert(elem != NULL);
	elem->Freeze();
      }

      Rectangle* rect = new Rectangle;
      getFrameBoundingBox(i(), LAYOUT_AUTO).ToRectangle(i()->GetX(), i()->GetY(), *rect);
      shapeFactory.Add(rect);
      if (i()->IsLast()) shapeFactory.SetNewRow();
    }
    shape = shapeFactory.GetShape();
  }
}

void
MathMLTokenElement::Render(const DrawingArea& area)
{
  if (!HasDirtyChildren()) return;

  RenderBackground(area);

  gc->SetColor(color);

  //area.DrawRectangle(gc, *shape);

  for (Iterator<MathMLFrame*> i(content); i.More(); i.Next()) {
    MathMLFrame* node = i();
    assert(node != NULL);
    if (node->IsText()) {
      MathMLTextNode* text = TO_TEXT(node);
      assert(text != NULL);
      text->Render(area);
    }
  }

  ResetDirty();
}

void
MathMLTokenElement::GetLinearBoundingBox(BoundingBox& b) const
{
  b.Zero();

  for (Iterator<MathMLFrame*> i(content); i.More(); i.Next()) {
    assert(i() != NULL);
    if (i()->IsText()) {
      MathMLTextNode* text = TO_TEXT(i());
      assert(text != NULL);
      const BoundingBox& textBox = text->GetBoundingBox();
      b.Append(textBox);
    }
  }
}

bool
MathMLTokenElement::IsToken() const
{
  return true;
}

bool
MathMLTokenElement::IsBreakable() const
{
  return true;
}

BreakId
MathMLTokenElement::GetBreakability() const
{
  // we have to skip some empty elements (malignmark) and get
  // the breakability of the last text node

  Iterator<MathMLFrame*> frame(content);
  frame.ResetLast();
  while (frame.More()) {
    assert(frame() != NULL);
    if (frame()->IsText()) break;
    frame.Prev();
  }

  return frame.More() ? frame()->GetBreakability() : BREAK_AUTO;
}

void
MathMLTokenElement::SetDirty(const Rectangle* rect)
{
  assert(IsShaped());

#if 0
  if (rect != NULL) {
    cout << NameOfTagId(IsA()) << " set dirty: shape " << *shape << " rect " << *rect;
    cout << " overlaps? " << shape->Overlaps(*rect) << endl;
  } else {
    cout << NameOfTagId(IsA()) << " set dirty!" << endl;
  }
#endif

#if 0
  cout << NameOfTagId(IsA()) << ": ";
  if (GetParent() != NULL)
    cout << GetParent()->IsDirty() << ' ' << GetParent()->IsSelected() << ' ' << IsSelected() << endl;
  else
    cout << "set dirty" << endl;
#endif

  dirtyBackground =
    (GetParent() != NULL && (GetParent()->IsSelected() != IsSelected())) ? 1 : 0;
#if 0
  if (GetParent() != NULL && (GetParent()->IsSelected() != IsSelected()))
    dirtyBackground = 1;
#endif

  if (IsDirty()) return;
  if (rect != NULL && !shape->Overlaps(*rect)) return;

  dirty = 1;
  SetDirtyChildren();

  for (Iterator<MathMLFrame*> frame(content); frame.More(); frame.Next()) {
    assert(frame() != NULL);
    frame()->SetDirty(rect);
  }
}

scaled
MathMLTokenElement::GetLeftEdge() const
{
  scaled edge = 0;
  bool first = true;

  for (Iterator<MathMLFrame*> frame(content); frame.More(); frame.Next()) {
    assert(frame() !=  NULL);
    if (first) {
      edge = frame()->GetLeftEdge();
      first = false;
    } else
      edge = scaledMin(edge, frame()->GetLeftEdge());
  }

  return edge;
}

scaled
MathMLTokenElement::GetRightEdge() const
{
  scaled edge = 0;
  bool first = true;

  for (Iterator<MathMLFrame*> frame(content); frame.More(); frame.Next()) {
    assert(frame() !=  NULL);
    if (first) {
      edge = frame()->GetRightEdge();
      first = false;
    } else
      edge = scaledMax(edge, frame()->GetRightEdge());
  }

  //  cout << NameOfTagId(IsA()) << " getting right edge: " << sp2ipx(edge) << endl;

  return edge;
}

scaled
MathMLTokenElement::GetDecimalPointEdge() const
{
  for (Iterator<MathMLFrame*> frame(content); frame.More(); frame.Next()) {
    assert(frame() != NULL);
    if (frame()->IsText()) {
      MathMLTextNode* text = TO_TEXT(frame());
      assert(text != NULL);
      if (text->HasDecimalPoint()) return text->GetDecimalPointEdge();
    }
  }

  return GetRightEdge();
}

bool
MathMLTokenElement::IsNonMarking() const
{
  for (Iterator<MathMLFrame*> frame(content); frame.More(); frame.Next()) {
    assert(frame() != NULL);
    if (frame()->IsText()) {
      MathMLTextNode* text = TO_TEXT(frame());
      assert(text != NULL);
      if (!text->IsNonMarking()) return false;
    }
  }

  return true;
}
