// Copyright (C) 2000, Luca Padovani <luca.padovani@cs.unibo.it>.
// 
// This file is part of GtkMathView, a Gtk widget for MathML.
// 
// GtkMathView is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// GtkMathView is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with GtkMathView; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// 
// For details, see the GtkMathView World-Wide-Web page,
// http://cs.unibo.it/~lpadovan/mml-widget, or send a mail to
// <luca.padovani@cs.unibo.it>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <assert.h>
#ifdef HAVE_IOSTREAM
#include <iostream>
#else
#include <stdio.h>
#endif

#include "globals.hh"
#include "stringAux.hh"
#include "MathMLParseFile.hh"
#include "OperatorDictionary.hh"

const String*
getAttribute(xmlNodePtr node, const char* attr)
{
  xmlChar* attrVal = xmlGetProp(node, XMLSTRING(attr));
  if (attrVal == NULL) return NULL;

  return allocString(attrVal);
}

//

OperatorDictionary::OperatorDictionary()
{
  firstDefaults = NULL;
  firstItem = NULL;
}

OperatorDictionary::~OperatorDictionary()
{
  Delete();
}

void
OperatorDictionary::Delete()
{
  const OperatorDefaults* p = firstDefaults;
  while (p != NULL) { 
    const OperatorDefaults* next = p->next;
    delete p;
    p = next;
  }

  const OperatorDictionaryItem* q = firstItem;
  while (q != NULL) {
    const OperatorDictionaryItem* next = q->next;
    delete q->name;
    delete q;
    q = next;
  }

  firstItem = NULL;
  firstDefaults = NULL;
}

void
OperatorDictionary::Message(const char* filename,
			    const char* message,
			    const char* param) const
{
  if (verbose) {
#ifdef HAVE_IOSTREAM
    cerr << "WARNING: malformed dictionary `" << filename << "': " << message << endl;
    // FIXME: is there a standard way to do the same as for the stdio library???
#else
    fprintf(stderr, "WARNING: malformed dictionary `%s': ", filename);
    fprintf(stderr, message, param);
    fprintf(stderr, "\n");
#endif
  }
}

bool
OperatorDictionary::Load(const char* filename)
{
  xmlDocPtr doc = MathMLParseFile(filename, true);
  if (doc == NULL) return false;

  xmlNodePtr root = xmlDocGetRootElement(doc);
  if (root == NULL) {
    xmlFreeDoc(doc);
    Message(filename, "parse error");
    return false;
  }

  if (xmlStrcmp(root->name, XMLSTRING("dictionary"))) {
    xmlFreeDoc(doc);
    Message(filename, "could not find dictionary root element");
    return false;
  }

  xmlNodePtr op = root->children;
  while (op != NULL) {
    if (!xmlIsBlankNode(op) && !xmlStrcmp(op->name, XMLSTRING("operator"))) {
      xmlChar* opName = xmlGetProp(op, XMLSTRING("name"));
      if (opName != NULL) {
	const String* opString = allocString(opName);
	OperatorDefaults* def = new OperatorDefaults;

	def->form  	   = getAttribute(op, "form");
	def->fence 	   = getAttribute(op, "fence");
	def->separator     = getAttribute(op, "separator");
	def->lspace 	   = getAttribute(op, "lspace");
	def->rspace 	   = getAttribute(op, "rspace");
	def->stretchy      = getAttribute(op, "stretchy");
	def->direction     = getAttribute(op, "direction");
	def->symmetric     = getAttribute(op, "symmetric");
	def->maxsize       = getAttribute(op, "maxsize");
	def->minsize       = getAttribute(op, "minsize");
	def->largeop       = getAttribute(op, "largeop");
	def->movablelimits = getAttribute(op, "movablelimits");
	def->accent        = getAttribute(op, "accent");

	const OperatorDefaults* defaults = AlreadyDefined(def);
	if (defaults == NULL) {
	  def->next = firstDefaults;
	  firstDefaults = def;
	  defaults = def;
	} else delete def;

	OperatorDictionaryItem* item = new OperatorDictionaryItem;
	item->name     = opString;
	item->defaults = defaults;
	item->next     = firstItem;

	firstItem = item;
      } else {
	Message(filename, "could not find operator name");
      }
    } else if (!xmlIsBlankNode(op)) {
      Message(filename, "unknown element `%s'", CSTRING(op->name));
    }

    op = op->next;
  }

  xmlFreeDoc(doc);

  return true;
}

const OperatorDefaults*
OperatorDictionary::AlreadyDefined(const OperatorDefaults* def) const
{
  const OperatorDefaults* p = firstDefaults;
  while (p != NULL && !p->Equal(def)) p = p->next;

  return p;
}

void
OperatorDictionary::Search(const String* opName,
			   const OperatorDefaults** prefix,
			   const OperatorDefaults** infix,
			   const OperatorDefaults** postfix) const
{
  assert(opName != NULL);
  assert(prefix != NULL && infix != NULL && postfix != NULL);

  *prefix = *infix = *postfix = NULL;

  const OperatorDictionaryItem* p = firstItem;
  while (p != NULL) {
    if (p->name->Equal(*opName)) {
      assert(p->defaults != NULL);
      if (p->defaults->form->Equal("prefix")) *prefix = p->defaults;
      else if (p->defaults->form->Equal("infix")) *infix = p->defaults;
      else if (p->defaults->form->Equal("postfix")) *postfix = p->defaults;
      else {
	assert(IMPOSSIBLE);
      }
    }
    p = p->next;
  }
}
