// Copyright (C) 2000, Luca Padovani <luca.padovani@cs.unibo.it>.
// 
// This file is part of GtkMathView, a Gtk widget for MathML.
// 
// GtkMathView is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// GtkMathView is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with GtkMathView; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// 
// For details, see the GtkMathView World-Wide-Web page,
// http://cs.unibo.it/~lpadovan/mml-widget, or send a mail to
// <luca.padovani@cs.unibo.it>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <assert.h>
#include <getopt.h>
#ifdef HAVE_IOSTREAM
#include <iostream>
#else
#include <stdio.h>
#endif
#include <stdlib.h>

#include "defs.h"
#include "guiGTK.h"
#include "Parser.hh"
#include "String.hh"
#include "globals.hh"
#include "EntitiesTable.hh"
#include "MathMLTokenElement.hh"

enum CommandLineOptionId {
  OPTION_VERSION = 256,
  OPTION_HELP,
  OPTION_STATS,
  OPTION_DUMP,
  OPTION_DUMP_ENTITIES,
  OPTION_ENTITY_WARNING,
  OPTION_VERBOSE
};

// global options
bool entityWarning = false;

static char appName[64];

extern void *parseMathMLFile(char *);

static void printVersion()
{
#ifdef HAVE_IOSTREAM
  cout << appName << " - written by Luca Padovani (C) 2000." << endl;
#ifdef DEBUG
  cout << "Compiled " << __DATE__ << ' ' << __TIME__ << endl;
#endif
#else
  printf("%s - written by Luca Padovani (C) 2000.\n", appName);
#ifdef DEBUG
  printf("Compiled %s %s\n", __DATE__, __TIME__);
#endif
#endif
  exit(0);
}

static void printHelp()
{
  static char* helpMsg = "\
Usage: viewer [options] file ...\n\n\
  -v, --version                 Output version information\n\
  -h, --help                    This small usage guide\n\
  --stats                       Some statistics about MathML and structures size\n\
  --dump                        Dump the file\n\
  --dump-entities               Dump a C table of the entities\n\
  --verbose                     Display warnings\n\
  --entity-warning              Warning for unknown entities\n\
";

#ifdef HAVE_IOSTREAM
  cout << helpMsg << endl;
#else
  printf("%s\n", helpMsg);
#endif

  exit(0);
}

static void printStats()
{
  printf("\
Class\t\tSize\n\
-----\t\t----\n\
Object\t\t%d\n\
Node\t\t%d\n\
Element\t\t%d\n\
Container\t\%d\n\
Value\t\t%d\n\
String\t\t%d\n\
\nCounters\n\
-------\n\
Values\t\t%d\n\
Strings\t\t%d\n\
Objects\t\t%d\n\
Elements\t%d\n\
Tokens\t\t%d\n\
\n",
	 sizeof(MathMLObject),
	 sizeof(MathMLNode),
	 sizeof(MathMLElement),
	 sizeof(MathMLContainerElement),
	 sizeof(Value),
	 sizeof(String),
#ifdef DEBUG
	 Value::GetCounter(),
	 String::GetCounter(),
	 MathMLObject::GetCounter(),
#else
	 0, 0, 0,
#endif
	 MathMLElement::GetCounter(),
	 MathMLTokenElement::GetCounter());
}

int main(int argc, char *argv[])
{
  sprintf(appName, "MathML Viewer v%d.%d.%d", VER_HI, VER_LO, VER_REV);

  bool dump = false;
  bool dumpEntities = false;
#ifdef MATHML_STATS
  bool viewStats = false;
#endif

  while (TRUE) {
    int option_index = 0;
    static struct option long_options[] =
    {
      { "version", 	  no_argument, NULL, OPTION_VERSION },
      { "help",    	  no_argument, NULL, OPTION_HELP },
      { "dump",    	  no_argument, NULL, OPTION_DUMP },
      { "dump-entities",  no_argument, NULL, OPTION_DUMP_ENTITIES },
      { "verbose",        no_argument, NULL, OPTION_VERBOSE },
      { "entity-warning", no_argument, NULL, OPTION_ENTITY_WARNING },
#ifdef MATHML_STATS
      { "stats",   no_argument, NULL, OPTION_STATS },
#endif
      { NULL,      no_argument, NULL, 0 }
    };

    int c = getopt_long(argc, argv, "vh", long_options, &option_index);

    if (c == -1) break;

    switch (c) {
    case OPTION_VERSION:
    case 'v':
      printVersion();
      break;

    case OPTION_HELP:
    case 'h':
      printHelp();
      break;

    case OPTION_DUMP:
      dump = true;
      break;

    case OPTION_DUMP_ENTITIES:
      dumpEntities = true;
      break;

    case OPTION_ENTITY_WARNING:
      entityWarning = true;
      break;

    case OPTION_VERBOSE:
      verbose = true;
      break;

#ifdef MATHML_STATS
    case OPTION_STATS:
      viewStats = true;
      break;
#endif

    case '?':
      break;

    default:
#ifdef HAVE_IOSTREAM
      cout << "*** getopt returned " << c << " value" << endl;
#else
      printf("*** getopt returned %o value\n", c);
#endif
      break;
    }
  }

  if (optind < argc) {
    GUI_init(&argc, &argv, appName, 500, 500);

    if (GUI_load_document(argv[optind]))
#ifdef HAVE_IOSTREAM
      cout << "viewer: fatal error: cannot load document `" << argv[optind] << '\'' << endl;
#else      
      printf("viewer: fatal error: cannot load document `%s'\n", argv[optind]);
#endif
    else {

      if (dump) GUI_dump_document();
      if (dumpEntities) entitiesTable.Dump();

      if (!viewStats) GUI_run();
      GUI_uninit();
      GUI_unload_document();
    }
  } else if (!viewStats) printHelp();

#ifdef MATHML_STATS
  if (viewStats) printStats();
#endif

  exit(0);
}
